class BinaryTreeNode<E> {
    private E data;
    private BinaryTreeNode<E> left;
    private BinaryTreeNode<E> right;

    public BinaryTreeNode(E data, BinaryTreeNode<E> left, BinaryTreeNode<E> right) {
        this.data = data;
        this.left = left;
        this.right = right;
    }
    // create leaf node: new BinaryTreeNode("6", null, null);

    public BinaryTreeNode(E data) {
        this(data, null, null); // call 3-argument constructor
    }
    // create a leaf node: new BinaryTreeNode("6");

    public E getData() {
        return data;
    }

    public BinaryTreeNode<E> getLeft() {
        return left;
    }

    public BinaryTreeNode<E> getRight() {
        return right;
    }

    public boolean isLeaf() {
        return left == null && right == null;
    }

    public int sum(BinaryTreeNode<Integer> node) {
        // base cases
        if (node == null) {
            return 0;
        }
        // the above base case covers this one
        // if (node.isLeaf()) {
        //     return node.data;
        // }
        // recursive case
        return node.data + sum(node.left) + sum(node.right);
    }

    public static int evaluate(BinaryTreeNode<String> expr) {
        // return the numerical result from evaluting expr
        // base case
        if (expr.isLeaf()) {
            return Integer.parseInt(expr.data);
        } else {
            // recursive case
            int left = evaluate(expr.left);
            int right = evaluate(expr.right);

            String op = expr.data;
            switch(op) {
                case "+":
                    return left + right;
                case "-":
                    return left - right;
                case "*":
                    return left * right;
                case "/":
                    return left / right;
            }
            throw new UnsupportedOperationException(op + " is not a recognized operator");
        }
    }

    public static int height(BinaryTreeNode<String> node) {
        if (node == null) {
            return 0;
        }
        if (node.left == null && node.right == null) {
            return 0;
        }
        return 1 + Math.max(height(node.left), height(node.right));
    }

    public static void printExpr(BinaryTreeNode<String> expr) {
        if (expr == null) {
            return;
        }
        if (expr.isLeaf()) {
            System.out.print(expr.data);
            return;
        }
        System.out.print("(");
        printExpr(expr.left);
        System.out.print(expr.data);
        printExpr(expr.right);
        System.out.print(")");
    }

    public static void main(String[] args) {
        // construct (5 + 9) * 2 + 3
        BinaryTreeNode<String> five = new BinaryTreeNode<>("5");
        BinaryTreeNode<String> nine = new BinaryTreeNode<>("9");
        BinaryTreeNode<String> fivePlusNine = new BinaryTreeNode<>("+", five, nine);
        BinaryTreeNode<String> fivePlusNineTimesTwo = new BinaryTreeNode<>("*", 
                        fivePlusNine, new BinaryTreeNode<String>("2"));
        BinaryTreeNode<String> fivePlusNineTimesTwoPlusThree = new BinaryTreeNode<>("+", 
                        new BinaryTreeNode<String>("3"), fivePlusNineTimesTwo);
        printExpr(fivePlusNineTimesTwoPlusThree);
        System.out.println(evaluate(fivePlusNineTimesTwoPlusThree));
    }
}